/**
* crype
* Copyright (c) 2014, The crype Authors
*
* Permission to use, copy, modify, and/or distribute this software for any purpose with
*  or without fee is hereby granted, provided that the above copyright notice and this
*  permission notice appear in all copies.
*/

/*

=IMPLEMENTATION GOALS=
 
  Encrypt data while still preserving good rsyncability.
  Must be usable through an unnamed shell pipe.
  Must work on Linux and Windows.

  Compression is implemented solely because it improves rsyncability compared
   to a separate gzip --rsyncable pass.

=LIBRARIES=

  Chacha20 stream cipher
  HMAC-SHA1-PBKDF2 hash functions
  Zlib-compatible compression

=FILE FORMAT=

  32-bit magic number "cry\x01"
  32-bit flags field (unused)

  Then follows an unspecified field for a plaintext header. This is currently
   unused, but the intention is it might include indexable nonsensitive metadata
   about the file's contents.
  32-bit length field
  {plaintext follows}

  Then follows a number of chunks until EOF. Each chunk takes the following
   format:
  32-bit marker
     -- Low 31 bits: Length of chunk
	                 (we know overflow will not occur because CRYPE_MAX_SINGLE_LENGTH)
	 -- High bit: Whether or not the chunk is zlib-compressed
  {chunk data follows}

=MEMBERS=

  crype::enckey is a 512-bit buffer.
  The first 256 bits are filled by SHA1-PBKDF2 from the supplied passphrase, and
   used as the chacha stream cipher key.
  The second 256 bits are set for every chunk, by HMAC-SHA1 of the first 4096
   bits of (compressed) plaintext. This is then truncated to 64 bits to use as
   the IV for the stream cipher.
  We know there will be at least 4096 bytes of (compressed) plaintext because of
   the CRYPE_MIN_SINGLE_LENGTH definition.

*/

#include "ccrype.h"

#include "vendor/pbkdf2_sha1/sha1.h"
#include "vendor/pbkdf2_sha1/sha1-pbkdf2.h"
#include "vendor/chacha/chacha.h"
#include "vendor/miniz/miniz.h"

#include <cstring> // memset

// n.b. There is an updated miniz available in valvesoftware/vogl with bugfixes / Zip64 support / Zip archive
//   support made optional, but it does not build cleanly and requires some further work to separate from vogl stuff.
// @ref http://richg42.blogspot.co.nz/2014/03/zip64-version-of-miniz-library-released.html

#define CRYPE_PBKDF2_ITERATIONS 4096

// Chunks to read from stdin at once. We should probably keep this below linux PIPE_BUF (default 4096) since that
//  is the largest amount guaranteed to be able to read atomically (does that matter for us?)
#define CRYPE_STDIN_READ_SIZE 8192

// Trigger if the sum of the past {CRYPE_DECISION_YAW} bytes equals zero mod {CRYPE_DECISION_PITCH}.
#define CRYPE_DECISION_YAW 256
#define CRYPE_DECISION_PITCH 4096

#define CRYPE_BUFFER_RESERVE ( 256 * CRYPE_DECISION_PITCH )

#define CRYPE_READ_HMAC_BYTES 4096

#define CRYPE_MIN_SINGLE_LENGTH 4096
#define CRYPE_MAX_SINGLE_LENGTH ( 16 * 1024 * 1024 )

#define CRYPE_CRYPT_USES_HMAC_BYTES 8

#if CRYPE_READ_HMAC_BYTES > CRYPE_MIN_SINGLE_LENGTH
#error "crype: MIN_SINGLE_LENGTH must be gte READ_HMAC_BYTES!"
#endif
#if CRYPE_CRYPT_USES_HMAC_BYTES > SHA1_MAC_LEN
#error "crype: USES_HMAC_BITS is greater than sha1 mac length!"
#endif

//

// cry\x01 - backwards for little-endian x86
#ifdef ECRYPT_LITTLE_ENDIAN
#define CRYPE_FH_MAGIC '\x01yrc'
#endif
#ifdef ECRYPT_BIG_ENDIAN
#define CRYPE_GH_MAGIC 'cry\x01'
#endif

struct crype_file_header {
	I32T magic;
	I32T flags;
	I32T desc_len;
};

//

crype::crype() 
	: in(stdin), out(stdout), total_chains(0), compression_level(MZ_DEFAULT_LEVEL)
{
	header.clear();
};


void crype::set_encryption_key(std::string& pass_str) {

	memset(enckey, 0, CRYPE_ENCKEY_MEMBER_SIZE);

	// Derive 512-bit key from passphrase into static global
	pbkdf2_sha1(pass_str.c_str(), pass_str.length(), CRYPE_PBKDF2_ITERATIONS, enckey, CRYPE_ENCKEY_MEMBER_SIZE);

	// The first 256 bits (32 bytes) are for encrypting data streams
	// The second 256 bits are for the HMAC key
}


void crype::set_compression_level(int compression_level) {
	this->compression_level = compression_level;
}


void crype::handle_enc(std::vector<unsigned char>& buff) {

	// Take HMAC of first CRYPE_READ_HMAC_BYTES bytes. We are guaranteed at least this
	//  many by CRYPE_MIN_SINGLE_LENGTH
	unsigned char hmac_out[SHA1_MAC_LEN] = { 0 };
	if (hmac_sha1(&enckey[CRYPE_ENCKEY_MEMBER_SIZE / 2], CRYPE_KEYSIZE_BITS / 8, &buff[0], CRYPE_READ_HMAC_BYTES, hmac_out) != 0) {
		fputs("crype: error calculating hmac", stderr);
		exit(1);
	}

	// Initialise chacha stream cipher
	ECRYPT_ctx ctx = { 0 };
	ECRYPT_keysetup(&ctx, enckey, CRYPE_KEYSIZE_BITS, SHA1_MAC_LEN*8);

	// MAC
	ECRYPT_ivsetup(&ctx, hmac_out); // chacha only uses the first 8 bytes of the 20-byte mac
	fwrite(hmac_out, CRYPE_CRYPT_USES_HMAC_BYTES, 1, out); // truncated 20->8

	// Encrypt buffer with chacha
	std::vector<unsigned char> crypted(buff.size());
	ECRYPT_encrypt_bytes(&ctx, &buff[0], &crypted[0], buff.size());

	// Echo output
	fwrite(&crypted[0], crypted.size(), 1, out);

}

void crype::handle_zip_enc(std::vector<unsigned char>& buff) {
	if (!buff.size()) {
		return; // fancy hat and nothing to say
	}

	bool done = false;

	if (compression_level) {

		// Compress buffer
		std::vector<unsigned char> compressed;
		compressed.resize(buff.size());
		unsigned long compressed_size = compressed.size();

		int did_compress = mz_compress2(&compressed[0], &compressed_size, &buff[0], buff.size(), compression_level);

		if (did_compress == MZ_OK) {
			compressed.resize(compressed_size); // truncate

			unsigned int marker = compressed.size() + CRYPE_CRYPT_USES_HMAC_BYTES;
			marker |= (1 << 31); // is_compressed=true
			fwrite(&marker, sizeof(unsigned int), 1, out);

			handle_enc(compressed);
			done = true;
		}
	}
	
	if (!done) {
		// compression disabled, failed, or didn't become any smaller than the input.
		// Just encrypt original data

		unsigned int marker = buff.size();
		fwrite(&marker, sizeof(unsigned int), 1, out);
		handle_enc(buff);

	}

	// Clear buff
	buff.clear();

	total_chains++;
}


bool crype::encrype() {

	unsigned int yaw_sum = 0;
	unsigned char yaw[CRYPE_DECISION_YAW] = { 0 };
	unsigned int yaw_pos = 0;

	std::vector<unsigned char> towrite;
	towrite.reserve(CRYPE_BUFFER_RESERVE);

	// write header

	crype_file_header cfh = { 0 };
	cfh.magic = CRYPE_FH_MAGIC;
	cfh.flags = 0;
	cfh.desc_len = header.length();

	fwrite(&cfh, sizeof(crype_file_header), 1, out);
	if (header.length()) {
		fwrite(&header[0], header.length(), 1, out);
	}

	// write chains

	for (;;) {
		// Check for eof/ferror on input
		if (feof(in) || ferror(in)) {
			break;
		}

		// Read from input
		unsigned char buff[CRYPE_STDIN_READ_SIZE] = { 0 };
		size_t read = fread(buff, 1, CRYPE_STDIN_READ_SIZE, in);
		if (!read) continue;

		for (size_t i = 0; i < read; ++i) {

			// Copy byte
			towrite.push_back(buff[i]);

			// Rolling sum of last _YAW bytes
			unsigned char yaw_last = yaw[yaw_pos];
			yaw[yaw_pos] = buff[i];
			yaw_pos = (yaw_pos + 1) % CRYPE_DECISION_YAW;
			yaw_sum = (yaw_sum + buff[i] - yaw_last) % CRYPE_DECISION_PITCH; // This might overflow negative(?)

			// Take decision on plaintext input
			if (yaw_sum == 0 && towrite.size() > CRYPE_MIN_SINGLE_LENGTH) {
				handle_zip_enc(towrite);
			}

		}

		// Force a chain if we havn't had one in ages
		if (towrite.size() > CRYPE_MAX_SINGLE_LENGTH) {
			handle_zip_enc(towrite);
		}
	}

	// Flush remaining buffers
	handle_zip_enc(towrite);

	return true;
}


void crype::handle_unzip(std::vector<unsigned char>& chain) {

	// FIXME: BAD ESTIMATE and lots of memory usage
	std::vector<unsigned char> uncompressed(chain.size() * 1024);

	unsigned long uncompressed_len = uncompressed.size();
	int did_uncompress = mz_uncompress(&uncompressed[0], &uncompressed_len, &chain[0], chain.size());
	if (did_uncompress != MZ_OK) {
		fprintf(stderr, "crype: failed to decompress! (%d)\n", did_uncompress);
		exit(1);

	}

	// Echo output
	fwrite(&uncompressed[0], uncompressed_len, 1, out);

}


void crype::handle_chain_unzip_enc(std::vector<unsigned char>& buff) {

	std::vector<unsigned char> decrypted(buff.size() - CRYPE_CRYPT_USES_HMAC_BYTES);

	// Set up chacha
	ECRYPT_ctx ctx = { 0 };
	ECRYPT_keysetup(&ctx, enckey, CRYPE_KEYSIZE_BITS, SHA1_MAC_LEN * 8);
	ECRYPT_ivsetup(&ctx, &buff[0]); // first CRYPT_USES_HMAC_BYTES=8 bytes are the truncated MAC, used as IV
	
	// Decrypt
	ECRYPT_encrypt_bytes(&ctx, &buff[CRYPE_CRYPT_USES_HMAC_BYTES], &decrypted[0], buff.size() - CRYPE_CRYPT_USES_HMAC_BYTES);

	// Pass to unzip function
	handle_unzip(decrypted);
}


bool crype::decrype(bool header_only) {

	// FIXME: it would be neat if this could work without allocating large amounts of
	//  memory at once, to protect against malicious input.
	// I know both miniz and chacha are capable of this
	// It's situationally unlikely anyone other than the encryptor will be decrypting

	// Read file header

	crype_file_header cfh = { 0 };
	if (fread(&cfh, sizeof(crype_file_header), 1, in) != 1) {
		fprintf(stderr, "crype: couldn't read header\n");
		return false;
	}
	if (cfh.magic != CRYPE_FH_MAGIC) {
		fprintf(stderr, "crype: not a crype file\n");
		return false;
	}

	// Read plaintext description

	header.clear();
	if (cfh.desc_len) {
		header.resize(cfh.desc_len);
		if (fread(&header[0], cfh.desc_len, 1, in) != 1) {
			fprintf(stderr, "crype: couldn't read header\n");
			return false;
		}
	}

	if (header_only) {
		return true;
	}

	// Process chains

	for (;;) {
		if (feof(in) || ferror(in)) {
			break;
		}

		// read marker
		unsigned int marker = 0;
		if (fread(&marker, sizeof(unsigned int), 1, in) != 1) { // reading 1 element
			if (feof(in)) {
				break; // apparently doesn't get caught earlier
			}

			fprintf(stderr, "crype: couldn't read marker\n");
			return false;
		}

		total_chains++;

		unsigned int is_compressed = (marker & (1 << 31));
		unsigned int zlen = marker & ~(1 << 31);

		// alloc zlen memory...
		std::vector<unsigned char> buff(zlen);

		// read data from pipe
		unsigned int remaining = zlen;
		do {
			unsigned int read_size = (remaining > CRYPE_STDIN_READ_SIZE) ? CRYPE_STDIN_READ_SIZE : remaining;
			unsigned int did_read = fread(&buff[zlen - remaining], read_size, 1, in);
			if (did_read != 1) { // reading 1 element
				fprintf(stderr, "crype: couldn't read enough data\n");
				return false;
			}
			remaining -= read_size;
		} while (remaining);

		// unzip/decrypt to stdout
		if (is_compressed) {
			handle_chain_unzip_enc(buff);
		} else {
			handle_enc(buff);
		}
	}

	return true;

}
