/**
 * crype
 * Copyright (c) 2014, The crype Authors
 *
 * Permission to use, copy, modify, and/or distribute this software for any purpose with
 *  or without fee is hereby granted, provided that the above copyright notice and this
 *  permission notice appear in all copies.
 */

/*
Test arguments:

-e -i input.dat -o output3.dat --password "asdf" --header "header comment1"
-d -i output3.dat -o output4.dat --password "asdf"
--read-header -i output3.dat

FIXME: something is causing different output each time release builds are recompiled
       We're going to need known test cases

*/


#include "ccrype.h"

#include <cstdlib> // atoi
#include <cstring> // strcmp
#include <fcntl.h> // _O_BINARY

#ifdef _WIN32
#include <io.h> // _setmode (MSVC + cygwin, unknown on debian)
#define SHOULD_SETMODE_BINARY
#else
#undef SHOULD_SETMODE_BINARY
#endif

enum crype_cli_mode {
	//CCM_UNKNOWN,
	CCM_ENCRYPT,
	CCM_DECRYPT,
	CCM_FILEHEADER
};

int usage() {
	fputs(
		"Usage: crype [options]\n"
		"\n"
		"  -e                      Mode: Encrypt (default)\n"
		"  -d                      Mode: Decrypt\n"
		"  --read-header           Mode: Read file header only\n"
		"\n"
		"  -i {file}               Set input from file (default stdin)\n"
		"  -o {file}               Set output to file (default stdout)\n"
		"  --password {str}        Set encryption password\n"
		"  --password-file {file}  Set encryption password from file\n"
		"  -v                      Verbose\n"
		"  --help                  Display this message\n"
		"  --no-password-warning   Suppress warning using password on command line\n"
		"\n"
		"Encoding options:\n"
		"  --header {str}          Set plaintext header\n"
		"  --header-file {file}    Set plaintext header from file\n"
		"  -c {level}              Set compression level (0-9)\n"
		,
		stderr
	);
	return 1;
}

bool crype_cli_read_file(const char* fname, std::string& dest) {
	FILE* pf = fopen(fname, "wb");
	if (pf == NULL) {
		return false;
	}

	fseek(pf, 0, SEEK_END);
	long pf_len = ftell(pf);
	dest.clear();
	dest.resize(pf_len);

	if (fread(&dest[0], pf_len, 1, pf) != 1) {
		return false;
	}

	fclose(pf);
	return true;
}

int main(int argc, char** argv) {
	
	std::string pass_str = "";
	enum crype_cli_mode climode = CCM_ENCRYPT;
	bool verbose = false;
	bool password_warning = true;

	crype c;

	if (argc == 1) {
		return usage();
	}

	for (int i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "-e")) {
			climode = CCM_ENCRYPT;

		} else if (!strcmp(argv[i], "-d")) {
			climode = CCM_DECRYPT;

		} else if (!strcmp(argv[i], "--read-header")) {
			climode = CCM_FILEHEADER;

		} else if (!strcmp(argv[i], "-i")) {
			c.in = fopen(argv[++i], "rb");
			if (c.in == NULL) {
				fprintf(stderr, "crype: Error opening \"%s\" for input\n", argv[i]);
				return 1;
			}

		} else if (!strcmp(argv[i], "-o")) {
			c.out = fopen(argv[++i], "wb");
			if (c.out == NULL) {
				fprintf(stderr, "crype: Error opening \"%s\" for output\n", argv[i]);
			}

		} else if (!strcmp(argv[i], "--no-password-warning")) {
			password_warning = false;

		} else if (!strcmp(argv[i], "--password")) {
			if (password_warning) {
				fputs("crype: warning: It's insecure to set passwords on the command-line!\n", stderr);
			}
			pass_str.assign(argv[++i]);

		} else if (!strcmp(argv[i], "--password-file")) {
			if (!crype_cli_read_file(argv[++i], pass_str)) {
				fprintf(stderr, "crype: Error reading from password file\n");
				return 1;
			}

		} else if (!strcmp(argv[i], "-c")) {
			const char* cval = argv[++i];
			if (strlen(cval) == 1 && cval[0] >= '0' && cval[0] <= '9') {
				c.set_compression_level(atoi(cval));
			}

		} else if (!strcmp(argv[i], "-v")) {
			verbose = true;

		} else if (!strcmp(argv[i], "--header")) {
			c.header = argv[++i];

		} else if (!strcmp(argv[i], "--header-file")) {
			if (!crype_cli_read_file(argv[++i], c.header)) {
				fprintf(stderr, "crype: Error reading from header file\n");
				return 1;
			}

		} else if (!strcmp(argv[i], "--help")) {
			return usage();

		} else {
			fprintf(stderr, "crype: Unrecognised option \"%s\"\n\n", argv[i]);
			return usage();

		}
	}

	if (!pass_str.length() && climode != CCM_FILEHEADER) {
		fputs("crype: No password specified\n", stderr);
		return 1;
	}

#ifdef SHOULD_SETMODE_BINARY
	if (c.in == stdin) {
		if (_setmode(_fileno(c.in), _O_BINARY) == -1) {
			fprintf(stderr, "crype: WARNING - couldn't set stdin to binary mode!\n");
		}
	}

	if (c.out == stdout) {
		if (_setmode(_fileno(c.out), _O_BINARY) == -1) {
			fprintf(stderr, "crype: WARNING - couldn't set stdout to binary mode!\n");
		}
	}
#endif


	c.set_encryption_key(pass_str);

	switch (climode) {
		case CCM_ENCRYPT: {
			c.encrype();
		} break;

		case CCM_DECRYPT: {
			c.decrype(false);
		} break;

		case CCM_FILEHEADER: {
			c.decrype(true);
			puts(c.header.c_str());
		} break;

		default:
			return usage();
	}

	if (verbose) {
		fprintf(stderr, "crype: Processed %d chain(s)\n", c.total_chains);
	}

	return 0;
}
